"""
We use a simple example to demonstrate the superiority of Wasserstein Distance over Pearson Correlation.
"""

import numpy as np
from scipy.stats import pearsonr
from scipy.stats import wasserstein_distance_nd
import matplotlib.pyplot as plt

np.random.seed(42)
real_data_x = np.random.normal(0, 1, 1000)
real_data_y = real_data_x + 10

# Generate Synthetic Data (Scaled Version)
synthetic_data_x = np.random.normal(-1, 1, 1000)
synthetic_data_y = synthetic_data_x + 100

# Compute Pearson Correlation
pearson_real = pearsonr(real_data_x, real_data_y)[0]
pearson_synthetic = pearsonr(synthetic_data_x, synthetic_data_y)[0]

print(f"Pearson Correlation (Real Data): {pearson_real}")
print(f"Pearson Correlation (Synthetic Data): {pearson_synthetic}")
print(f"Pearson Correlation Difference: {pearson_synthetic - pearson_real}")

# Compute Wasserstein Distance for both dimensions
real_data = np.column_stack((real_data_x, real_data_y))
synthetic_data = np.column_stack((synthetic_data_x, synthetic_data_y))
wassertein = wasserstein_distance_nd(real_data, synthetic_data)

# Print Results

print(f"Wasserstein Distance: {wassertein}")
